import "Equendil.Utils.ByteStream";
import "Equendil.Utils.Zlib";
import "Equendil.Utils.bit32";
import "Equendil.Utils.Log";
import "Equendil.Utils.TurbineUTF8Binary";

LICodec = {};

function LICodec.CanEncode( item )
	if item:GetID() then
		return true;
	else
		return false;
	end
end


local log2 = function( x )
	return math.log( x ) / math.log( 2 );
end

-- Single-precision floating-point format  - IEEE 754-2008 
--  TODO:handle numbers out of bound
local numberToBinary32Float = function( n )
	if n == 0 then return 0 end;

	local precision = 24;
	local sign;
	if n > 0 then
		sign = 0;
	else 
		sign = 1;
	end
	n = math.abs( n );

	local maxn = 2^(precision - 1)
	local exponent = math.ceil(log2(math.floor(n)));
	
	if n > (2*maxn) then
		while n > maxn do
			n = n / 2;
		end
	else
		while n < maxn do
			n = n * 2;
			if n < 0 then
				exponent = exponent - 1;
			end
		end
	end
	
	-- bias the exponent
	exponent = exponent + 127
	
	local result = sign;
	result = Equendil.Utils.bit32.lshift( result, 8 ); -- exponent is encoded on 8bits
	result = result + exponent;
	result = Equendil.Utils.bit32.lshift( result, 23 ); -- and significand on 23 bits
	result = result + math.floor( n - 2^precision ); -- we don't store the most significant bit (always 1, as number is normalized)
	return result;
end

-- converts UTF8 to UTF16
-- only the basic multilingual plane is supported
local UTF8toUTF16 = function( name )
	local encodedArray = {};
	local string_byte = string.byte;
	local table_insert = table.insert;
	local bit_extract = Equendil.Utils.bit32.extract;
	local byte;

	local i = 1;
	while i <= #name do
		byte = string_byte( name, i, i );
		if bit_extract( byte, 5, 3 ) == 6 then
			-- two bytes UTF sequence: 110xxxxx 10xxxxxx
			i = i + 1;
			local byte2 = string_byte( name, i, i );
			local out = ( bit_extract( byte, 0, 5) * 64 ) + bit_extract( byte2, 0, 6 );
		
			table_insert( encodedArray, out );
		elseif bit_extract( byte, 4, 4 ) == 14 then
			-- three bytes UTF sequence:  1110xxxx 10xxxxxx 10xxxxxx
			i = i + 1;
			local byte2 = string_byte( name, i, i );
			i = i + 1;
			local byte3 = string_byte( name, i, i );
			local out = ((( bit_extract( byte, 0, 4 ) * 64 ) + bit_extract( byte2, 0, 6 ) ) * 64) + bit_extract( byte3, 0, 6 );
			table_insert( encodedArray, out );
		else
			-- either a byte or a 4+ bytes sequence, which we don't support,, either way, it's left as is
			table_insert( encodedArray, byte );
		end
		i = i + 1;
	end
	
	return encodedArray;
end

function LICodec.Serialize( item )
	local out = Equendil.Utils.ByteStream();
	
	-- Instance ID
	out:PutLongLE( math.random(0,0x7FFFFFFF) );
	out:PutLongLE( 0x03000000 );
	
	-- Generic ID
	local id = item:GetID();
	-- id = nil;
	out:PutLongLE( id or 0x700302Eb );
	
	-- has a name
	out:Put( 0x01 ); 
	
	-- name length
	local name = UTF8toUTF16( item:GetName() );
	name[256] = nil; -- truncate if necessary
	out:Put( #name );
	-- name
	for _,c in ipairs( name ) do
		out:PutWordLE( c );
	end
	
	-- unknown fields
	out:PutLongLE( 0x00000001 );
	out:PutWordLE( 0x0001 );
	
	-- LI Title
--	out:PutLongLE( 0x7002CE9D );
	out:PutLongLE( 0 );
	
	-- Unknown field
	out:Put( 0x00 );
	
	-- legacies
	local nLegs = 0;
	for _, legacy in pairs( item:GetLegacies() ) do
		if legacy:IsActive() and ( not legacy:IsDefault() ) then
			nLegs = nLegs + 1;
		end
	end

	if nLegs > 20 then nLegs = 20 end;
	out:Put( nLegs );
	local n = 1;
	for _, legacy in ipairs( item:GetLegacies() ) do
		if nLegs == 0 then 
			break;
		end
		if legacy:IsActive() and ( not legacy:IsDefault() ) then
			local id = legacy:GetID();
			if id == nil then
				out:PutLongLE( 0x7002C011 ); -- generic legacy
			else
				out:PutLongLE( id ); -- id
			end
			out:PutLongLE( legacy:GetInternalRank() ); 
			-- out:PutLongLE( ranks[n] );			-- rank
			n = n +1;
			nLegs = nLegs - 1;
		end
	end

	-- Relics
	out:Put( 0x01 ); -- no relics
	out:Put( 0x00 ); -- 0 relics

	-- Stats
	out:PutLongLE( 0 ); -- 0 stats
	
	-- Points spent and left
	out:PutLongLE( item:GetTotalPoints() - item:GetSpentPoints() );
	out:PutLongLE( item:GetSpentPoints() );

	-- DPS Rank
	local defaultLegacy = item:GetDefaultLegacy();
	if defaultLegacy:GetType() == Type.DPS then
		-- dps rank
		out:PutLongLE( defaultLegacy:GetInternalRank() ); -- dps rank
		out:PutLongLE( 0 ); -- default legacy rank
		out:PutLongLE( 0 ); -- default legacy ID
	else
		out:PutLongLE( 0 ); -- dps rank
		out:PutLongLE( defaultLegacy:GetInternalRank() or 0 ); -- default legacy rank
		out:PutLongLE( defaultLegacy:GetID() or 0 ); -- default legacy ID
	end

	-- extra info
	out:PutLongLE( 0x100010EF );
	out:Put(0);
	out:Put(3); -- 3 extra structures
	
	-- extra extra info
	out:PutLongLE( 0x100012C5 );	-- header
	out:PutLongLE( 0x100012C5 );	-- header repeated
	
	-- number of extra extra structures
	local extras = 6;
	if item:GetCrystals() > 0 then
		extras = extras + 1;
	end
	
	local tactWeaponDPS = item:GetTacticalWeaponDPS();
	if tactWeaponDPS ~= nil then
		extras = extras + 1;
	end
	
	-- if defaultLegacy:GetType() ~= Type.DPS then
		-- extras = extras + 1;
	-- end
	
	out:PutLongLE( extras );
	
	-- if defaultLegacy:GetType() ~= Type.DPS then
		-- default legacy rank
		-- out:PutLongLE( 0x100031A4 );	-- header
		-- out:PutLongLE( defaultLegacy:GetInternalRank() or 0 );
	-- end
	
	-- second value of dps range in floating point (IEEE 754)
	if tactWeaponDPS ~= nil then
		out:PutLongLE( 0x10001042 );	-- header
		local maxHit = item:GetTacticalWeaponDPS() * item:GetAttackDuration() * (5/4);
		out:PutLongLE( numberToBinary32Float( maxHit ) );
	end
	
	-- storage
	-- out:PutLongLE( 0x0000034E );
	-- out:PutLongLE( 0x20010000 );

	-- craft name
	out:PutLongLE( 0x10000884 );	-- header
	out:Put( 1 );					-- bool flag of some sort ?
	out:Put( 4 );					-- name length
	out:PutLongLE( 0x00610046 );	-- "Fa"
	out:PutLongLE( 0x0065006B );	-- "ke"
	out:PutLongLE( 0x00000001 );	-- unknown field
	out:PutWordLE( 0x0000 );		-- unknown field

	-- crafted by
	out:PutLongLE( 0x10000E20 );	-- header
	out:Put( 1 );					-- bool flag of some sort ?
	out:Put( 31 );					-- name length
	local name = {0x0065004C,0x00650067,0x0064006e,0x00720061,0x00200079,0x00740049,0x006d0065,0x00500020,0x0061006c,0x006e006e,0x00720065,0x00280020,0x006c0050,0x00670075,0x006e0069}
	for _,v in ipairs( name ) do
		out:PutLongLE( v); 
	end
	out:PutWordLE( 0x0029 );		-- ")"
	out:PutWordLE( 1 );				-- unknown field
	out:PutLongLE( 0 );				-- unknown field

	-- bound to
	-- out:PutLongLE(0x10000AC1); 
	-- out:PutLongLE(0x0000E09E);
	-- out:PutLongLE(0x02160000);	
	
	-- LI Level
	out:PutLongLE( 0x10001D5F ); 	-- header
	out:PutLongLE( item:GetLevel() );	-- level

	-- item level
	out:PutLongLE( 0x100000C4 );	-- header
	out:PutLongLE( item:GetLevelToUse() );
	
	-- durability
	out:PutLongLE( 0x1000132C );
	out:PutLongLE( 0x00000001 );
	
	-- vlevel
	out:PutLongLE( 0x10000669 );
	local vlevel = VirtualLevel( item.levelToUse, item.age, item.crystals, item.rank )
	-- out:PutLongLE( 109 );
	out:PutLongLE( vlevel );

	-- upgrades
	if item:GetCrystals() > 0 then
		out:PutLongLE( 0x10004996 );	-- header
		out:PutLongLE( item:GetCrystals() );	-- #crystals
	end
	
	-- GenericID
	out:PutLongLE( 0x10000421 );	-- header
	out:PutLongLE( 0x10000421 );	-- header repeated
	out:PutLongLE( id or 0x700302Eb );
	
	-- Instance ID
	out:PutLongLE( 0x10002897 );	-- header
	out:PutLongLE( 0x10002897 );	-- header repeated
	out:PutLongLE( 0x00000000 );	-- instance id low
	out:PutLongLE( 0x03160003 );	-- instance id high
	
	return out:GetData();
end

function LICodec.Encode( item )
	local table_insert = table.insert;

	-- serialize the item
	local serializedData = LICodec.Serialize( item );
	
	-- compress it
	local deflatedData = Equendil.Utils.Zlib.Deflate( serializedData );
	
	-- array with header + data. Header stats with a 0 longword
	local preEncodingArray = {0,0,0,0};
	
	-- append length of (inflated) data to header
	local len = #serializedData;
	for i = 1,4 do
		local byte = len % 256;
		table_insert( preEncodingArray, byte );
		len = Equendil.Utils.bit32.rshift( len, 8 );
	end
	
	-- append data
	for _,byte in ipairs( deflatedData ) do
		table_insert( preEncodingArray, byte );
	end
	
	-- encode it to Uncode characters using Turbine's scheme
	local encodedString = Equendil.Utils.TurbineUTF8Binary.Encode( preEncodingArray );
	
	return encodedString;
end

local HasLegacy = function( liData, legacyID )
	for _,legacy in ipairs( liData.legacies ) do
		if legacy.IDs then
			for _,id in ipairs( legacy.IDs ) do
				if id == legacyID then
					return true;
				end
			end
		end
	end
	return false;
end

local FindLIData = function( legacyID )
	if HasLegacy( BridleLight, legacyID ) then
		return 1,nil;
	elseif HasLegacy( BridleMedium, legacyID ) then
		return 2,nil;
	elseif HasLegacy( BridleHeavy, legacyID ) then
		return 3,nil;
	end
	
	for _,class in pairs( ClassIndices ) do
		local legendaryItems = ClassesInfo[class].legendaryItems;
		for i = 1,(#legendaryItems - 3) do -- -3 to skip bridles
			if HasLegacy( legendaryItems[i], legacyID ) then
				return i,class;
			end
		end
	end
	return nil;
end

local DecodeName = function( byteStream )
	local nameLen = byteStream:Get();
	local nameArray = {};
	for i = 1, nameLen do
		table.insert( nameArray, byteStream:GetWordLE() );
	end
	local name = string.char( unpack( nameArray ) );
	return name;
end

function LICodec.Decode( data, name )
	-- Unencode and uncompress the data
	local decodedTable = Equendil.Utils.TurbineUTF8Binary.Decode( data );
	local deflatedDataAsString = string.char( unpack( decodedTable ) );
	deflatedDataAsString = string.sub( deflatedDataAsString, - ( string.len( deflatedDataAsString ) - 8 ) );
	local inflatedData = Equendil.Utils.Zlib.Inflate( deflatedDataAsString );

	-- set up the data as a 'stream', sort of
	local ins = Equendil.Utils.ByteStream();
	ins:SetData( inflatedData );
	
-- Retrieve what info we can
	
	-- Instance ID
	ins:Consume(8);
	
	-- Generic ID
	local GID = ins:GetLongLE();
	
	-- item name
	local fName = ins:Get();
	local itemName;
	if fName == 1 then
		itemName = DecodeName( ins );
	else
		ins:Consume(8);	-- localization id(s)
	end

	-- ??
	ins:Consume(6); 
	
	-- LI title
	ins:Consume(4); 
	
	-- legacies
	ins:Consume(1); -- ??

	local nLegs = ins:Get();
	if nLegs == 0 then
		-- Unidentified LIs, we'll have too little information to determine class, level etc, so we return in error
		-- Equendil.Utils.Log.Error( "LIP", "Item Decoding: No legacies found, cannot determine LI" );
		return nil;	
	end

	local legacies = {};
	for i = 1, nLegs do
		local ID = ins:GetLongLE();
		local rank = ins:GetLongLE();
		table.insert( legacies, { ["ID"] = ID, ["rank"] = rank } );
	end
	
	-- relics
	local fRelics = ins:Get();
	local nRelics = ins:Get();
	ins:Consume( 8 * nRelics );
	
	-- item stat IDs, if we have 3 or more, assume it's a two handed weapons
	local nStats = ins:GetLongLE();
	ins:Consume( 4 * nStats );
	local isTwoHanded = ( nStats >= 3 );
	
	-- Pinpoint the LI we're dealing with
	local liIndex, classIndex = FindLIData( legacies[1].ID );
	if liIndex == nil then
		-- not found
		-- Equendil.Utils.Log.Error( "LIP", "Item Decoding: couldn't determine the LI type" );
		return nil;
	elseif classIndex == nil then
		-- bridle
		-- Equendil.Utils.Log.Error( "LIP", "Item Decoding: Bridles unsupported at this time" );
		return nil;
	end

	-- FindLIData will return the first data set that matches,  
	-- which may be a guardian or champion 1h weapon, or a captain 2h weapon,
	-- adjust accordingly if we have a two handed guardian or champion weapon, or one handed captain weapon
	local class = ClassesInfo[classIndex];
	if isTwoHanded then
		if 	( classIndex == ClassIndices[Turbine.Gameplay.Class.Guardian] ) or
			( classIndex == ClassIndices[Turbine.Gameplay.Class.Guardian] ) then
			-- two handed weapon for a guardian or champion, 
			liIndex = 2;
		end
	else 
		if ( classIndex == ClassIndices[Turbine.Gameplay.Class.Captain] ) then
			-- one handed weapon for a captain
			liIndex = 2;
		end
	end

	-- create a legendary item with what we've found
	local item = Equendil.LIP.LegendaryItem( class.legendaryItems[liIndex] );
		
	-- points spent / left
	local pointsSpent = ins:GetLongLE();
	local pointsLeft = ins:GetLongLE();
	
	-- default legacy
	local dpsRank = ins:GetLongLE();
	local defLegID = ins:GetLongLE();
	local defLegRank = ins:GetLongLE();

	-- extra info as an array 
	local header = ins:GetLongLE();
	if header ~= 0x100010EF then
		Equendil.Utils.Log.Error( "LIP", "Item Decoding: missing 100010EF header" );
		return nil;
	end
	local fLevelSet = false;
	
	ins:Consume(1);
	local nSubs = ins:Get();
	for i = 1, nSubs do 
		header = ins:GetLongLE();
		header2 = ins:GetLongLE();
		if header == nil then
			Equendil.Utils.Log.Error( "LIP", string.format( "Item decoding: Top level header is nil" ) );
		elseif header ~= header2 then
			Equendil.Utils.Log.Error( "LIP", string.format( "Item decoding: Top level headers not matching: %08x - %08x" ) );
			return nil;
		elseif header == 0x100012C5 then 
			-- extra extra info as another array therein...
			local nExtras = ins:GetLongLE();
			for i = 1, nExtras do 
				local header = ins:GetLongLE();
				if header == 0x10000E20 then	
					-- crafted by ... 
					ins:Consume(1);
					local nameLen = ins:Get();
					ins:Consume( nameLen * 2 );
					ins:Consume( 6 );
				elseif header == 0x10000AC1 then
					-- bound to 
					ins:Consume( 8 );
				elseif header == 0x10001D5F then
					-- LI level
					local level = ins:GetLongLE();
					item:SetLevel( level );
				elseif header == 0x100031A4 then
					-- default legacy rank
					ins:Consume( 4 );
				elseif header == 0x100026BC then
					-- legacies, we have them already
					local nLegs = ins:GetLongLE();
					ins:Consume( 8 * nLegs );
				elseif header == 0x10000669 then
					-- virtual level
					ins:Consume( 4 );
				elseif header == 0x10000884 then
					-- crafting name
					ins:Consume(1);
					local nameLen = ins:Get();
					ins:Consume( nameLen * 2 );
					ins:Consume( 6);
				elseif header == 0x100000C4 then
					-- item level
					local level = ins:GetLongLE( 4 );
					item:SetLevelToUse( level );
					fLevelSet = true;
				elseif header == 0x10004996 then
					-- upgrades (crystals)
					local crystals = ins:GetLongLE();
					item:SetCrystals( 3 );
				elseif header == 0x100038A7 or 	-- unknown 
						header == 0x10000AC2 	-- unknown 
				then
					ins:Consume( 1 );
				elseif 	header == 0x1000132C or	-- durability
						header == 0x10000835 or	-- worth
						header == 0x10000E7B or	-- quantity
						header == 0x0000034E or -- storage info
						header == 0x10000ACD or	-- unknown
						header == 0x10001042	-- unknown
				then
					ins:Consume(4);
				end
			end
		elseif header == 0x10000421 then
			-- GID
			ins:Consume(4);
		elseif header == 0x10002897 then
			-- IID
			ins:Consume(8);
		end
	end

	local fAgeSet = true;
	-- guess the age of the LI
	if string.find( name, L["First Age"] ) then
		item:SetAge( 1 );
	elseif string.find( name, L["Second Age"] ) then
		item:SetAge( 2 );
	elseif string.find( name, L["Third Age"] ) then
		item:SetAge( 3 );
	else
		item:SetAge( 3 );
		fAgeSet = false;
	end
	
	-- set values we know
	item:SetName( itemName or name );
	for _,legacyData in ipairs( legacies ) do
		local legacy, tier = item:FindLegacyByID( legacyData.ID );
		if legacy ~= nil then
			legacy:SetTier( tier );
			legacy:SetActive( true );
			legacy:SetInternalRank( legacyData.rank, fLeveLSet and fAgeSet );
		end
	end

	local defaultLegacy = item:GetDefaultLegacy();
	if defLegID ~= nil then
		defaultLegacy:SetInternalRank( defLegRank );
	else
		defaultLegacy:SetInternalRank( dpsRank );
	end
	
	return item, classIndex, liIndex;
end