--[[
ScrollBar class. Replacement for Turbine's ScrollBar that lacks a method to set the relative size of the slider (it uses the 'large change' value instead, which is inconvenient)  
Interface is the same as Turbine's but for the following differences:
	- orientation is to be set on construction and currently cannot be changed on the fly (SetOrientation() is unimplemented)
	- Two additional methods:
		- SetSliderProportion( proportion) sets the relative size of the slider (proportion between 0 and 1)
		- GetSliderProportion() returns that value

SetDecrementButton(), SetIncrementButton() & SetThumbButton() are not implemented (neither do they seem to be in Turbine's API)
--]]

import "Equendil.UI.IconButton";
import "Equendil.Utils.Event";

-- definitions and resources within game files

-- buttons, 10x10, normal, highlighted, depressed (same as normal)
local ArrowUpTiny = { 0x4100028C, 0x4100028D, 0x4100028C };
local ArrowDownTiny = { 0x4100028E, 0x4100028F, 0x4100028E };
local ArrowLeftTiny = { 0x41000290, 0x41000291, 0x41000290 };
local ArrowRightTiny = { 0x41000292, 0x41000293, 0x41000292 };

-- horizontal & vertical sliders, 10x10, normal, highlghted
local SliderH = { 
	CapBeg = { 0x41000286, 0x41000287 },
	Bar = { 0x41000288, 0x41000289 },
	CapEnd = { 0x4100028A, 0x4100028B }
}

local SliderV = {
	CapBeg  = { 0x41000280, 0x41000281 },
	Bar = { 0x41000282, 0x41000283 },
	CapEnd = { 0x41000284, 0x41000285 },
}

local Normal = 1;
local Highlighted = 2;

-- utility function that reverse coordinates depending on orientation
local function orientate( orientation, x, y )
	if orientation == Turbine.UI.Orientation.Horizontal then
		return x, y;
	else 
		return y, x;
	end
end

-- Slider class (no -/+ buttons)
local Slider = class( Turbine.UI.Control );
function Slider:Constructor( orientation )
	Turbine.UI.Control.Constructor( self );

	self.orientation = orientation;
	self.dragging = false;
	self.mouseIn = false;
	
	self.Dragged = Equendil.Utils.Event();
	
	self.capBeg = Turbine.UI.Control();
	self.capBeg:SetParent( self );
	self.capBeg:SetSize( 10, 10 );
	self.capBeg:SetBlendMode( Turbine.UI.BlendMode.Overlay );
	self.capBeg:SetMouseVisible( false );
	
	self.bar = Turbine.UI.Control();
	self.bar:SetParent( self );
	self.bar:SetBlendMode( Turbine.UI.BlendMode.Overlay );
	self.bar:SetMouseVisible( false );
	
	self.capEnd = Turbine.UI.Control();
	self.capEnd:SetParent( self );
	self.capEnd:SetSize( 10, 10 );
	self.capEnd:SetBlendMode( Turbine.UI.BlendMode.Overlay );
	self.capEnd:SetMouseVisible( false );
	
	if self.orientation == Turbine.UI.Orientation.Horizontal then
		self.resources = SliderH;
	else
		self.resources = SliderV;
	end
	
	self.capBeg:SetBackground( self.resources.CapBeg[Normal] );
	self.bar:SetBackground( self.resources.Bar[Normal] );
	self.capEnd:SetBackground( self.resources.CapEnd[Normal] );

	self:SetMouseVisible( true );
	self.MouseEnter = function() self:_MouseIn() end
	self.MouseLeave = function() self:_MouseOut() end
	self.MouseDown = function( _, args ) self:_BeginDrag( args ) end
	self.MouseMove = function( _, args ) self:_Drag( args ) end
	self.MouseUp = function() self:_EndDrag() end
	
	self:_DoLayout();
end

function Slider:_Orientate( x, y )
	return orientate( self.orientation, x, y );
end

function Slider:_MouseIn()
	self.mouseIn = true;
	if not self.dragging then
		self:_Highlight( true );
	end
end

function Slider:_MouseOut()
	self.mouseIn = false;
	if not self.dragging then
		self:_Highlight( false );
	end
end

function Slider:_Highlight( flag )
	local index = function() 
		if flag then 
			return Highlighted; 
		else 
			return Normal;
		end 
	end

	self.capBeg:SetBackground( self.resources.CapBeg[index()] );
	self.bar:SetBackground( self.resources.Bar[index()] );
	self.capEnd:SetBackground( self.resources.CapEnd[index()] );
end

function Slider:_BeginDrag( pos )
	self.dragging = true;
	
	self.orig = self:_Orientate( pos.X, pos.Y );
end

function Slider:_Drag( pos )
	if not self.dragging then return end
	
	local offset = self:_Orientate( pos.X, pos.Y ) - self.orig;
	self.Dragged( offset );
end

function Slider:_EndDrag()
	self.dragging = false;
	self:_Highlight( self.mouseIn );
end

function Slider:SetSize( width, height )
	Turbine.UI.Control.SetSize( self, width, height );
	self:_DoLayout();
end

function Slider:SetWidth( width )
	Turbine.UI.Control.SetWidth( self, width );
	self:_DoLayout();
end

function Slider:SetHeight( height )
	Turbine.UI.Control.SetHeight( self, height );
	self:_DoLayout();
end

function Slider:_DoLayout()
	local main, secondary = self:_Orientate( self:GetSize() );
	self.bar:SetPosition( self:_Orientate( 10, 0 ) );
	self.bar:SetSize( self:_Orientate( main - 20, secondary ) );
	self.capEnd:SetPosition( self:_Orientate( main - 10, 0 ) );
end

-- ScrollBar class (slider + button)
ScrollBar = class( Turbine.UI.Control ); -- we derive from the button class because Control() won't catch mouse wheel events for some reason
function ScrollBar:Constructor( orientation )
	Turbine.UI.Control.Constructor( self );
	
	self.OrientationChanged = Equendil.Utils.Event();
	self.ValueChanged = Equendil.Utils.Event();
	
	self.orientation = orientation;

	self.minValue = 1;
	self.maxValue = 1;
	self.smallChange = 1;
	self.largeChange = 1;
	self.value = 1;
	self.sliderProportion = 1;
	
	if orientation == Turbine.UI.Orientation.Horizontal then
		self.buttonDecrement = Equendil.UI.IconButton( ArrowLeftTiny, 10, 10 );
		self.buttonIncrement = Equendil.UI.IconButton( ArrowRightTiny, 10, 10 );
	else
		self.buttonDecrement = Equendil.UI.IconButton( ArrowUpTiny, 10, 10 );
		self.buttonIncrement = Equendil.UI.IconButton( ArrowDownTiny, 10, 10 );
	end

	self.buttonDecrement:SetParent( self );
	self.buttonIncrement:SetParent( self );
	
	self.slider = Slider( self.orientation );
	self.slider:SetParent( self );
	self.slider.Dragged = function( offset ) 
		-- compute & set the new value of the scrollbar
		local range = self.maxValue - self.minValue; -- range of possible values
		local containerLength = self:_Orientate( self:GetSize() ) - 20; -- Length of the container area
		local sliderLength = self.sliderProportion * containerLength ; -- length of the slider
		local sliderPos = self:_Orientate( self.slider:GetPosition() ) - 10; -- current position of the slider relative to its containing area
		local value = math.floor( ( sliderPos + offset ) * range / ( containerLength - sliderLength ) + self.minValue ); -- compute the new value from the position of the slider + offset
		value = Equendil.Utils.Maths.constrain( self.minValue, value, self.maxValue ); -- make sure the new values is within min ... max value
		self:SetValue( value );
	end
	
	self.buttonDecrement.Clicked = function() 
		self:SetValue( math.max( self.minValue, self.value - 1 ) );
	end
	
	self.buttonIncrement.Clicked = function()
		self:SetValue( math.min( self.maxValue, self.value + 1 ) );
	end
end


-- Scrollbar methods

function ScrollBar:GetDecrementButton() return self.buttonDecrement end
function ScrollBar:GetIncrementButton() return self.buttonIncrement end
function ScrollBar:GetThumbButton() return self.slider end
function ScrollBar:GetLargeChange()	return self.largeChange end
function ScrollBar:GetMaximum() return self.maxValue end
function ScrollBar:GetMinimum() return self.minValue end
function ScrollBar:GetOrientation() return self.orientation end
function ScrollBar:GetSmallChange() return self.smallChange end
function ScrollBar:GetValue() return self.value end
function ScrollBar:SetLargeChange(value) self.largeChange = value end
function ScrollBar:SetMaximum(value) self.maxValue = value; self:_DoLayout() end
function ScrollBar:SetMinimum(value) self.minValue = value; self:_DoLayout() end
function ScrollBar:SetSmallChange(value) self.smallChange = value end

function ScrollBar:SetValue(value)
	if value == self.value then return end
	
	self.value = value;
	self:_DoLayout();
	
	self.ValueChanged();
end

-- Unimplemented methods
function ScrollBar:SetDecrementButton(value) end
function ScrollBar:SetIncrementButton(value) end
function ScrollBar:SetThumbButton(value) end
function ScrollBar:SetOrientation(value) end


-- Additional methods
function ScrollBar:SetSliderProportion( proportion ) self.sliderProportion = proportion; self:_DoLayout(); end
function ScrollBar:GetSliderProportion() return self.sliderProportion; end

-- Overloaded control methods

function ScrollBar:SetWidth( width )
	Turbine.UI.Control.SetWidth( self, width );
	self:_DoLayout();
end

function ScrollBar:SetHeight( height )
	Turbine.UI.Control.SetHeight( self, height );
	self:_DoLayout();
end

function ScrollBar:SetSize( width, height )
	Turbine.UI.Control.SetSize( self, width, height );
	self:_DoLayout();
end

-- Private implementation

function ScrollBar:_Orientate( x, y )
	return orientate( self.orientation, x, y );
end

function ScrollBar:_DoLayout()
	local main, secondary = self:_Orientate( self:GetSize() );
	self.buttonIncrement:SetPosition( self:_Orientate( main - 10, 0 ) );
	
	local range = self.maxValue - self.minValue;

	local sliderLength = self.sliderProportion * ( main - 20 );
	self.slider:SetSize( self:_Orientate( sliderLength, 10 ) );
	
	local sliderPos = ( self.value - self.minValue ) / range * ( main - 20 - sliderLength ) + 10;
	self.slider:SetPosition( self:_Orientate( sliderPos, 0 ) );
	
	-- Turbine.Log:Information( "range : " .. tostring( range ) .. ", slider length : " .. tostring( sliderLength ) .. ", sliderPos : " .. tostring( sliderPos ) );
end


