import "Turbine.UI";

-- A VirtualListBox is a list box that doesn't handle its own data, instead requesting controls that need to be displayed on the fly.
-- This presents two important advantages
-- 1) Decoupling of data & user interface. UI elements shouldn't be handling data that is not immediately relevant to display or user interaction.
-- 2) No overhead in handling large lists, ie it's fast

-- NOTE: If selection mode is not set to NoSelection, events on item controls might be captured by this control
-- NOTE: MouseWheel event of the item controls *will* be captured by this control
-- NOTE: selection highlightning is handled by calling the SetSelected( true/false) method on an item
-- NOTE: Does not handle variable height items (pain in the bottom).. Set the item height !

-- TODO: multiple selection entirely untested
-- TODO: horizontal scrollbar

SelectionMode = { SingleSelection = 1, MultipleSelection = 2, NoSelection = 3 };

VirtualListBox = class( Turbine.UI.Button ); -- we derive from the button class because Control() won't catch mouse wheel events for some reason
function VirtualListBox:Constructor( getItemFunction, itemHeight )
	Turbine.UI.Button.Constructor( self );

	-- GetItemControl( n, selectionStatus ) is called for each item the listbox needs to draw.
	-- It should return a control for the nth item in the list
	-- selectionStatus = true if the item is selected, false or nil otherwise
	self.GetItemControl = getItemFunction; 

	-- event when selectiont has changed, no paramaters
	self.SelectionChanged = Equendil.Utils.Event();
	
	self.itemCount = 0;
	self.position = 1; -- position in the list
	self.numItemsDisplayed = 0;
	self.multiSelection = false;
	self.selectionMode = SelectionMode.SingleSelection;
	self.itemHeight = itemHeight;
	
	self.selectedIndices = {};
	self.lastSelectedIndex = -1;

	self.verticalScrollBar = nil;
	
	self.MouseWheel = function( sender, args )
		self:Scroll( - args.Direction );
	end
	
	self.actionValueChanged = function()
		local newPosition = self.verticalScrollBar:GetValue();
		if self.position ~= newPosition then
			self.position = newPosition;
			self:Refresh();
		end
	end
	
	self.itemControls = {};
end

function VirtualListBox:Scroll( shift )
	local maxPosition = self:_MaxPosition( self.itemCount );
	self.position = Equendil.Utils.Maths.constrain( 1, self.position + shift, maxPosition );
	self:_UpdateVerticalScrollBarPosition();
	self:Refresh();
end

function VirtualListBox:SetHeight( height )
	Turbine.UI.Button.SetHeight( self, height );
	self.numItemsDisplayed = math.floor( self:GetHeight() / self.itemHeight );
	
	-- move the current position if resizing makes the current position too low
	if self.position + self.numItemsDisplayed - 1 > self.itemCount then
		self.position = math.max( 1, self.itemCount - self.numItemsDisplayed + 1 );
		self:_UpdateVerticalScrollBarPosition();
	end
	self:_UpdateScrollBars();
	self:Refresh();
end
	
function VirtualListBox:SetSize( width, height )
	self:SetWidth( width );
	self:SetHeight( height );
end

-- set the height of any item that will be displayed in the list box, we do not handle variable height items
function VirtualListBox:SetItemHeight( itemHeight )
	self.itemHeight = itemHeight;
	self.numItemsDisplayed = math.floor( self:GetHeight() /  self.itemHeight );
end

-- Set the selection mode of the list box, one of the SelectionMode enum
function VirtualListBox:SetSelectionMode( selectionMode )
	self.selectionMode = selectionMode;
	
	if self.selectionMode ~= SelectionMode.MultipleSelection then
		self.selectedIndices = {};
	end
	
	if self.selectionMode == SelectionMode.NoSelection then
		self.lastSelectedIndex = -1;
	end
end

-- get the selection mode of the listbox
function VirtualListBox:GetSelectionMode()
	return self.selectionMode;
end

-- get the position within the list
function VirtualListBox:GetListPosition()
	return self.position;
end

-- set the position within the list
function VirtualListBox:SetListPosition( position )
	self.position = position;
	self:_UpdateVerticalScrollBarPosition();
	self:Refresh();
end

-- Ensure the nth item is visible in the listbox. If not visible it is brought to the top or bottom of the list depending on current position
function VirtualListBox:MakeVisible( n )
	if n < self.position then
		-- position is before the first item displayed, bring to top
		self.position = n; 
		self:Refresh();
	elseif n > self.position + self.numItemsDisplayed - 1 then
		-- position is after the last item displayed, bring it to bottom
		self.position = self:_MaxPosition( n );
		
		self:Refresh();
	else
		return; -- already visible
	end

	-- update the scroll bar accordingly
	self:_UpdateVerticalScrollBarPosition();
end

-- find what the position should be set so that the nth item is at the bottom. If there are too few items to display position is top of the list
function VirtualListBox:_MaxPosition( n )
	return math.max( 1, n - self.numItemsDisplayed + 1 );
end

-- set how many items are in the listbox, will cause loss of selection and refresh
function VirtualListBox:SetItemCount( n )
	self.itemCount = n;
	local maxPosition = self:_MaxPosition( n );
	if self.position > maxPosition then
		self.position = maxPosition;
	end
	-- get rid of old selection
	self.lastSelectedIndex = -1;
	self.selectedIndices = {};
	
	self:Refresh();
	self:_UpdateScrollBars();
	self:_UpdateVerticalScrollBarPosition();
	self.SelectionChanged();
end

-- Return the number of items in the listbox
function VirtualListBox:GetItemCount( n )
	return self.itemCount;
end

-- Clear all controls currently displayed
function VirtualListBox:_ClearItems()
	-- clear the item control list
	for i, control in ipairs( self.itemControls ) do
		control:SetParent( nil );
		-- clear callbacks if necessary
		if self.selectionMode ~= SelectionMode.NoSelection then
			control.MouseUp = nil; -- TODO: we might have zombie callbacks here if the selection mode is changed after initialization of the listbox
		end
	end
	self.itemControls = {};
end

-- Redraw the list
function VirtualListBox:Refresh()
	self:_ClearItems();
	
	-- get & position controls
	local top = 0;
	for i = self.position, self.itemCount do
		-- get the control to display
		local control = nil;
		if self.selectionMode == SelectionMode.MultipleSelection then
			control = self.GetItemControl( i, self.selectedIndices[i] );
		elseif self.selectionMode == SelectionMode.SingleSelection then
			control = self.GetItemControl( i, self.lastSelectedIndex == i );
		end
		
		-- and set it up
		control:SetParent( self );
		control:SetPosition( 0, top );
		control:SetMouseVisible( true );

		control.MouseWheel = function( sender, args  ) -- we have to capture mouse wheel events of the controls, we'll never get them ourselves
			self:Scroll( -args.Direction );
		end
		if self.selectionMode ~= SelectionMode.NoSelection then
			control.MouseUp = function( sender, args )
				if self.selectionMode == SelectionMode.MultipleSelection then
					self:_ToggleSelection( i );
				else
					self:Select( i );
				end
			end
		end

		table.insert( self.itemControls, control );
		
		-- if we're beyond the bottom, stop displaying controls
		top = top + control:GetHeight();
		if top >= self:GetHeight() then
			break;
		end
	end
end

-- Clear any and all selection
function VirtualListBox:ClearSelection()
	self.lastSelectedIndex = -1;
	self.selectedIndices = {};
end

-- returns the first selected index
function VirtualListBox:GetSelectedIndex()
	return self.lastSelectedIndex;
end

-- returns a table of all selected indices
function VirtualListBox:GetSelectedIndices()
	return self.selectedIndices;
end

-- returns the nth item to display provided it's visible (ie, being currently handled)
function VirtualListBox:_GetNthItem( n )
	if n >= self.position and n < self.position + self.numItemsDisplayed then
		return self.itemControls[n - self.position + 1];
	end
	
	return nil;
end

-- Set selection of the nth item to <flag>. If flag is nil then toggle the selection of that item.
-- External method
function VirtualListBox:Select( n )
	if self.selectionMode == SelectionMode.NoSelection then	return end

	if self.selectionMode == SelectionMode.MultipleSelection then
		self.selectedIndices[n] = true;
	else 
		local oldSelectedItem = self:_GetNthItem( self.lastSelectedIndex );
		if oldSelectedItem ~= nil then
			oldSelectedItem:SetSelected( false );
		end
	end

	self.lastSelectedIndex = n;
	local selectedItem = self:_GetNthItem(n);
	if selectedItem ~= nil then
		selectedItem:SetSelected( true );
	end
		
	self.SelectionChanged();
end

-- Toggle the selection of an item
-- Private method to react to user actions
function VirtualListBox:_ToggleSelection( n )
	if self.selectionMode ~= SelectionMode.MultipleSelection then return end

	if self.selectedIndices[n] then
		self.selectedIndices[n] = nil;
		self:_GetNthItem(n):SetSelected( false );
	else
		self.selectedIndices[n] = true;
		self:_GetNthItem(n):SetSelected( true );
	end

	-- if we've just deselected the last selected index, set it to the first selected index if any
	if self.lastSelectedIndex == n then
		self.lastSelectedIndex = -1;
		for i, v in ipairs(self.selectedIndices) do
			self.lastSelectedIndex = i;
			break;
		end
	end
end

-- Update the configuration of the scroll bars
function VirtualListBox:_UpdateScrollBars()
	-- if all the items are displayed make the scrollbar invisible
	if self.verticalScrollBar ~= nil then
		if self.numItemsDisplayed >= self.itemCount then
			self.verticalScrollBar:SetVisible( false );
		else
			self.verticalScrollBar:SetVisible( true );
			self.verticalScrollBar:SetMaximum( self:_MaxPosition( self.itemCount ) );
			self.verticalScrollBar:SetLargeChange( self.numItemsDisplayed );
			self.verticalScrollBar:SetSliderProportion( self.numItemsDisplayed / self.itemCount );
		end
	end
	
	 -- TODO: no horizontal scroll bar for now.
end


function VirtualListBox:_UpdateVerticalScrollBarPosition()
	if self.verticalScrollBar ~= nil then
		-- remove the event handler before we set the value, we don't want self generated events
		self.verticalScrollBar.ValueChanged:Remove( self.actionValueChanged );
		self.verticalScrollBar:SetValue( self.position );
		self.verticalScrollBar.ValueChanged:Add( self.actionValueChanged );
	end
end

-- Set a vertical scroll bar to the listbox
function VirtualListBox:SetVerticalScrollBar( scrollbar )
	self.verticalScrollBar = scrollbar;
	if self.verticalScrollBar == nil then return end
	
	self.verticalScrollBar:SetMinimum( 1 );
	self.verticalScrollBar:SetSmallChange( 1 );

	self:_UpdateScrollBars();
	
	self.verticalScrollBar.ValueChanged:Add( self.actionValueChanged );
end

-- Set an horizontal scroll bar to the listbox
function VirtualListBox:SetHorizontalScrollBar( scrollbar )
	-- TODO
end
